/*

Launcher stub for tlmgr gui (with hidden console window) with UAC prompt

This version written in 2014 and 2016 by Siep Kroonenberg, Public Domain
*/

#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0600
// #define NTDDI_VISTA // not needed after all
#include <windows.h>
#include <winnt.h>
#include <shellapi.h>
#include <shlobj.h>
#include <wtypes.h>
#include <wchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static wchar_t selfdir[MAX_PATH];
static char release[5];
static BOOL is_vista;

#define CMD_TAIL L"\\tlmgr.bat\" -gui"
// extra wchar for initial double quote
#define MAX_CMD (MAX_PATH + wcslen( CMD_TAIL ) + 1)

#define DIEFLAGS MB_ICONERROR | MB_TASKMODAL | MB_SETFOREGROUND
void die( wchar_t * s ) {
  MessageBox( NULL, s, L"ERROR!", DIEFLAGS );
  PostQuitMessage( 1 );
}

// functions for checking OS version
// adapted from http://www.codeproject.com/Articles/678606/
//     Part1-Overcoming-Windows-8-1s-deprecation-of-GetVe
BOOL geq_MajorVersion( DWORD majorVersion ) {
  OSVERSIONINFOEX osVersionInfo;
  ULONGLONG maskCondition;
  ZeroMemory( &osVersionInfo, sizeof( OSVERSIONINFOEX ));
  osVersionInfo.dwOSVersionInfoSize = sizeof( OSVERSIONINFOEX );
  osVersionInfo.dwMajorVersion = majorVersion;
  maskCondition = VerSetConditionMask( 0, VER_MAJORVERSION, VER_GREATER_EQUAL );
  return VerifyVersionInfo( &osVersionInfo, VER_MAJORVERSION, maskCondition );
}
BOOL geq_MinorVersion(DWORD minorVersion) {
  OSVERSIONINFOEX osVersionInfo;
  ULONGLONG maskCondition;
  ZeroMemory( &osVersionInfo, sizeof( OSVERSIONINFOEX ));
  osVersionInfo.dwOSVersionInfoSize = sizeof( OSVERSIONINFOEX );
  osVersionInfo.dwMinorVersion = minorVersion;
  maskCondition = VerSetConditionMask( 0, VER_MINORVERSION, VER_GREATER_EQUAL );
  return VerifyVersionInfo( &osVersionInfo, VER_MINORVERSION, maskCondition );
}

// >= Vista: check process token for elevation
// XP: check membership of administrator group

BOOL is_elevated() {

  BOOL ret = FALSE;

  if ( is_vista ) {
    // http://stackoverflow.com/questions/8046097/
    HANDLE ProcessHandle, TokenHandle = NULL;
    TOKEN_ELEVATION Elevation;
    DWORD cbsize = sizeof( TOKEN_ELEVATION );
    ProcessHandle = GetCurrentProcess(); // no need to close afterwards
    if( OpenProcessToken( ProcessHandle, TOKEN_QUERY, &TokenHandle )) {
      if( GetTokenInformation(
          TokenHandle,
          TokenElevation, // NOT a variable!
          &Elevation,
          sizeof( Elevation ),
          &cbsize )) {
        ret = Elevation.TokenIsElevated;
        if( TokenHandle ) CloseHandle( TokenHandle );
      }
    }
    return ret;
  } else { // <= xp
    BOOL b;
    SID_IDENTIFIER_AUTHORITY NtAuthority = {SECURITY_NT_AUTHORITY};
    PSID AdministratorsGroup;
    b = AllocateAndInitializeSid(
        &NtAuthority,
        2,
        SECURITY_BUILTIN_DOMAIN_RID, // system domain
        DOMAIN_ALIAS_RID_ADMINS,
        0, 0, 0, 0, 0, 0,
        &AdministratorsGroup
    );

    if ( b ) {
      if (! CheckTokenMembership( NULL, AdministratorsGroup, &b )) {
        b = FALSE;
      }
      FreeSid( AdministratorsGroup );
    }
    return b;
  }
}

void get_tl_release() {
  // read release from release-texlive.txt in root of installation
  wchar_t dbfilename[MAX_PATH];
  wchar_t * wptr;
  char * cptr;
  int i, j, l;
  FILE * dbfile;
  char dbline[200];

  // calculate filename of release file
  wcscpy( dbfilename, selfdir );
  for ( i=0; i<2; i++ ) {
    wptr = wcsrchr( dbfilename, L'\\' );
    if ( ! wptr ) die( L"No directory; should not have happened" );
    *wptr = L'\0';
  }
  if ( wcslen( dbfilename ) + wcslen( L"\\tlpkg\\texlive.tlpdb" ) >= MAX_PATH )
    die( L"Path too long" );
  wcscat( dbfilename, L"\\tlpkg\\texlive.tlpdb" );

  // read start of release file
  dbfile = _wfopen( dbfilename, L"r" );
  if ( !dbfile ) {
    release[0] = '\0';
    return;
  }
  for ( i=0; i<200; i++ ) {
    memset( dbline, 0, sizeof( dbline ));
    cptr = fgets( dbline, sizeof( dbline ), dbfile );
    if( !cptr ) {
      // i.e. EOF or error
      fclose( dbfile );
      release[0] = '\0';
      return;
    }
    l = strlen( "depend release/" );
    if( !strncmp( "depend release/", dbline, l)) {
      // next 4 chars should be digits
      for( j=0; j<4; j++ ) {
        release[j] = dbline[l+j];
        if( release[j] < '0' || release[j] > '9' ) {
          memset( release, 0, sizeof( release ));
          break;
        }
      }
      release[4] = '\0';
      fclose( dbfile );
      return;
    }
  }
  // if we got here, release should still consist
  // of zeroes from the initial default initialization
  fclose( dbfile );
  return;
}

BOOL in_HKLM() {
  /*
  Check for uninstall reg entry in HKLM for current TL.

  On w64, registry entries under HKLM\software\wow6432node will be
  transparently aliased to corresponding entries under HKLM\software
  so for 32-bits programs so no worries on this score.

  For just checking the existence of the uninstall registry key,
  we can get away with ascii/ansi/utf8.

  */
  long result;
  HKEY uninst_key;
  char keyname[] =
     "Software\\Microsoft\\windows\\currentversion\\uninstall\\TeXLivexxxx";

  if (!release[0]) return FALSE;
  strncpy( keyname + strlen( keyname ) - 4, release, 4 );

  result = RegOpenKeyExA(
    HKEY_LOCAL_MACHINE,
    keyname,
    0,
    KEY_QUERY_VALUE,
    &uninst_key);
  // An additional check: compare the path of the uninstaller
  // with the path of this program, using
  // the GetFileInformationByHandle API call, this time using wchar_t
  return ( result == ERROR_SUCCESS );
}

// starting the real tlmgr
void start_real_tlmgr() {
  wchar_t cmdln[ MAX_CMD ];
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  // compose command-line
  cmdln[ 0 ] = L'\0';
  wcscat( cmdln, L"\"" );
  wcscat( cmdln, selfdir );
  wcscat( cmdln, CMD_TAIL );
  // create child process the normal way
  ZeroMemory( &si, sizeof(si) );
  si.cb = sizeof(si);
  si.dwFlags = STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE ;
  ZeroMemory( &pi, sizeof(pi) );
  if( !CreateProcess(
    NULL,     // module name (uses command line if NULL)
    cmdln,    // command line
    NULL,     // process security atrributes
    NULL,     // thread security atrributes
    TRUE,     // handle inheritance
    0,        // creation flags, e.g. CREATE_NEW_CONSOLE, CREATE_NO_WINDOW, DETACH
    NULL,     // pointer to environment block (uses parent if NULL)
    NULL,     // starting directory (uses parent if NULL)
    &si,      // STARTUPINFO structure
    &pi )     // PROCESS_INFORMATION structure
  ) die( L"Failed to start TL manager" );
  else PostQuitMessage( 0 );
}

// starting the real tlmgr with an elevation prompt
void start_real_tlmgr_elevated() {
  wchar_t cmdln[ MAX_CMD ];
  SHELLEXECUTEINFO shExecInfo;
  // compose full filename of batchfile
  cmdln[ 0 ] = L'\0';
  wcscat( cmdln, L"\"" );
  wcscat( cmdln, selfdir );
  wcscat( cmdln, L"\\tlmgr.bat\"" );
  // this is the recommended way to get an elevation prompt
  shExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
  shExecInfo.fMask = SEE_MASK_NO_CONSOLE;
  shExecInfo.hwnd = NULL;
  shExecInfo.lpVerb = L"runas";
  shExecInfo.lpFile = cmdln;
  shExecInfo.lpParameters = L"-gui";
  shExecInfo.lpDirectory = NULL;
  shExecInfo.nShow = SW_HIDE;
  shExecInfo.hInstApp = NULL;
  if ( !ShellExecuteEx( &shExecInfo ))
    die( L"Failed to start TL manager elevated" );
  else PostQuitMessage( 0 );
}

int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
    PWSTR cmdline, int iShow ) {
  DWORD l;
  wchar_t * wptr;

  // windows version check
  is_vista = geq_MajorVersion( 6 );
  if ( ! geq_MajorVersion( 5 ))
    die( L"XP minimum Windows version" );
  if (( ! geq_MajorVersion( 6 )) && ( ! geq_MinorVersion( 1 )))
    die( L"XP minimum Windows version for GUI TL Manager" );

  // calculate own directory selfdir
    // msdn is conflicted as to whether _wpgmptr is absolute;
    // don't use _get_wpgmptr, although GetModuleFileName locks out w2k
  l = GetModuleFileName( NULL, selfdir, MAX_PATH ); // length l excludes \0
  if ( l >= MAX_PATH ) die( L"Cannot get own filename" );
  wptr = wcsrchr( selfdir, L'\\' );
  if ( !wptr ) die( L"No directory; should not have happened" );
  *wptr = L'\0';

  // read TL release from release file
  get_tl_release();
  if( ! release[0] ) {
    MessageBox( NULL, L"Problem with installation: cannot determine release!",
        L"Error", MB_OK | MB_ICONSTOP );
    return 1;
  }
  // check for an uninstall entry under HKLM
  if ( ! in_HKLM() )
    start_real_tlmgr();
  else if ( is_elevated() )
    start_real_tlmgr();
  else if ( is_vista )
    start_real_tlmgr_elevated();
  else
    MessageBox( NULL, L"Only administrators may run this program!",
        L"Error", MB_OK | MB_ICONSTOP );
  return 0;
}
